import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        let renderer = new THREE.WebGLRenderer({ antialias: true });
        this.domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        scene.background = new THREE.Color(0x050505);
        this._scene = scene;

        this._setupCamera();
        this._setupModel();
        this._setupGUI();
        
        window.onresize = this.resize.bind(this);

        this.resize();
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(30, window.innerWidth / window.innerHeight, 1, 10000);
        camera.position.z = 300;
        this._camera = camera;
        
    }

    _createShaderMaterial() {
        const uniforms = {
            "amplitude": { value: 1.0 },
            "color": { value: new THREE.Color(0xff2200) },
            "colorTexture": { value: new THREE.TextureLoader().load("textures/water.jpg") }
        };

        uniforms["colorTexture"].value.wrapS = uniforms["colorTexture"].value.wrapT = THREE.RepeatWrapping;

        const vertexCode = `
            uniform float amplitude;
            attribute float displacement;
            varying vec3 vNormal;
            varying vec2 vUv;

            void main() {
                vNormal = normal;
                vUv = (0.5 + amplitude) * uv + vec2(amplitude);
                vec3 newPosition = position + amplitude * normal * vec3(displacement);
                gl_Position = projectionMatrix * modelViewMatrix * vec4(newPosition, 1.0);
            }
        `;

        const fragmentCode = `
            varying vec3 vNormal;
            varying vec2 vUv;
            uniform vec3 color;
            uniform sampler2D colorTexture;

            void main() {
                vec3 light = vec3(0.5, 0.2, 1.0);
                light = normalize(light);

                float dProd = dot(vNormal, light) * 0.5 + 0.5;

                vec4 tcolor = texture2D(colorTexture, vUv);
                vec4 gray = vec4(vec3(tcolor.r*0.3 + tcolor.g*0.59 + tcolor.b*0.11), 1.0);
                gl_FragColor = gray * vec4(vec3(dProd) * vec3(color), 1.0);
            }
        `;
        
        const shaderMaterial = new THREE.ShaderMaterial({
            uniforms: uniforms,
            vertexShader: vertexCode,
            fragmentShader: fragmentCode

        });

        return { material: shaderMaterial, uniforms: uniforms };
    }

    _setupModel() {
        const radius = 50, segments = 128, rings = 64;
        const geometry = new THREE.SphereGeometry(radius, segments, rings);

        const displacement = new Float32Array(geometry.attributes.position.count);
        const noise = new Float32Array(geometry.attributes.position.count);

        for(let i=0; i<displacement.length; i++) {
            noise[i] = Math.random() * 5;
        }

        geometry.setAttribute("displacement", new THREE.BufferAttribute(displacement, 1));

        const material = this._createShaderMaterial();
        const sphere = new THREE.Mesh(geometry, material.material);
        this._scene.add(sphere);

        this._sphere = sphere;
        this._uniforms = material.uniforms;
        this._displacement = displacement;
        this._noise = noise;
    }

    update() {
        const time = Date.now() * 0.01;
        this._sphere.rotation.y = this._sphere.rotation.z = 0.01 * time;
        this._uniforms["amplitude"].value = 2.5 * Math.sin(this._sphere.rotation.y * 0.125);
        this._uniforms["color"].value.offsetHSL(0.0005, 0, 0);

        for(let i=0; i<this._displacement.length; i++) {
            this._displacement[i] = Math.sin(0.1 * i + time);
            this._noise[i] += 0.5 * (0.5 - Math.random());
            this._noise[i] = THREE.MathUtils.clamp(this._noise[i], -5, 5);
            this._displacement[i] += this._noise[i];
        }

        this._sphere.geometry.attributes.displacement.needsUpdate = true;
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
        this._stats.update();
    }

    resize() {
        this._camera.aspect = window.innerWidth / window.innerHeight;
        this._camera.updateProjectionMatrix();

        this._renderer.setPixelRatio(window.devicePixelRatio);
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }

    _setupGUI() {
        const stats = new Stats();
        this.domWebGL.appendChild(stats.dom);
        this._stats = stats;
    }
}

window.onload = function () {
    (new App()).render();
}